import sys
import os
import pickle
import random
from copy import deepcopy

from transformers import BertModel, BertForTokenClassification, BertTokenizer

from parc_loader import load_corpus

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

import numpy as np

from crf_layer import CRFOutputLayer
import evaluation

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class BertWrapper(nn.Module):
    def __init__(self, num_labels=None):
        super().__init__()
        self.bert = BertForTokenClassification.from_pretrained('bert-base-uncased', num_labels=num_labels)
        
        self.num_labels = self.bert.config.num_labels
        self.max_length = self.bert.config.max_position_embeddings

    def loss(self, inputs, labels):
        sequence_length = inputs.shape[1]
        if sequence_length <= self.max_length:
            return self.bert(inputs, labels=labels)[0]
        # The whole thing with gradients doesn't fit in memory
        # just pick one window to compute loss for
        window_start = random.choice(list(range(0, sequence_length, self.max_length//2)))
        input_window = inputs[:,window_start:window_start+self.max_length].contiguous()
        label_window = labels[:,window_start:window_start+self.max_length].contiguous()
        window_length = input_window.shape[1]
        # input_window: float32[batch_size, window_length]
        return self.bert(input_window, labels=label_window)[0]
        
    def forward(self, inputs):
        sequence_length = inputs.shape[1]
        if sequence_length <= self.max_length:
            logits = self.bert(inputs)[0]
        else:
            output_windows = []
            for window_start in range(0, sequence_length, self.max_length//2):
                input_window = inputs[:,window_start:window_start+self.max_length]
                window_length = input_window.shape[1]
                # input_window: float32[batch_size, window_length]
                output_window = self.bert(input_window)[0]
                # output_window: float32[batch_size, window_length]
                output_windows.append(output_window)
            window_start_index = self.max_length//4
            window_end_index = 3*self.max_length//4
            parts = []
            parts.append(output_windows[0][:,:window_end_index])
            for output_window in output_windows[1:-1]:
                parts.append(output_window[:,window_start_index:window_end_index])
            parts.append(output_windows[-1][:,window_start_index:])
            logits = torch.cat(parts, dim=1)
        #logits: float32[batch_size, sequence_length, n_labels]
        n_labels = logits.shape[-1]
        predictions = torch.argmax(logits, dim=-1)
        # predictions: int64[batch_size, max_length]
        return predictions

class BertCRF(nn.Module):
    def __init__(self, n_labels):
        super().__init__()
        self.bert = BertModel.from_pretrained('bert-base-uncased')
        self.n_units = self.bert.config.hidden_size
        self.n_labels = n_labels
        self.max_length = self.bert.config.max_position_embeddings
        self.crf = CRFOutputLayer(self.n_units, self.n_labels)

    def forward(self, inputs):
        sequence_length = inputs.shape[1]
        if sequence_length <= self.max_length:
            bert_outputs = self.bert(inputs)[0]
        else:
            output_windows = []
            for window_start in range(0, sequence_length, self.max_length//2):
                input_window = inputs[:,window_start:window_start+self.max_length]
                window_length = input_window.shape[1]
                # input_window: float32[batch_size, window_length]
                output_window = self.bert(input_window)[0]
                # output_window: float32[batch_size, window_length]
                output_windows.append(output_window)
            window_start_index = self.max_length//4
            window_end_index = 3*self.max_length//4
            parts = []
            parts.append(output_windows[0][:,:window_end_index])
            for output_window in output_windows[1:-1]:
                parts.append(output_window[:,window_start_index:window_end_index])
            parts.append(output_windows[-1][:,window_start_index:])
            bert_outputs = torch.cat(parts, dim=1)
        #bert_outputs: float32[batch_size, sequence_length, self.n_units]
        return self.crf(bert_outputs)

    def loss(self, inputs, labels):
        sequence_length = inputs.shape[1]
        if sequence_length <= self.max_length:
            bert_outputs = self.bert(inputs)[0]
        else:
            # The whole thing with gradients doesn't fit in memory
            # just pick one window to compute loss for
            window_start = random.choice(list(range(0, sequence_length, self.max_length//2)))
            input_window = inputs[:,window_start:window_start+self.max_length].contiguous()
            label_window = labels[:,window_start:window_start+self.max_length].contiguous()
            window_length = input_window.shape[1]
            # input_window: float32[batch_size, window_length]
            bert_outputs = self.bert(input_window)[0]
            labels = label_window
        return self.crf.loss(bert_outputs, labels, None)



class BertSequenceLabelingModel:
    def __init__(self, model_name, architecture, batch_size, log=print):
        self.model_name = model_name
        self.use_crf = (architecture == 'bert-crf')
        self.batch_size = batch_size
        self.log = log
        self.initialized = False
        self.label_indices = None
        self.label_indices_t = None
        self.tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')

    def save(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        d = dict(self.__dict__)
        d['bert'] = None
        del d['log']
        del d['model_name']
        with open(path, 'wb') as f:
            pickle.dump(d, f)
            if self.initialized:
                torch.save(self.bert.state_dict(), f)

    def load(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        with open(path, 'rb') as f:
            d = pickle.load(f)
            for k, v in d.items():
                self.__dict__[k] = v
            if self.initialized:
                if self.use_crf:
                    self.bert = BertCRF(len(self.label_indices)).to(device)
                else:
                    self.bert = BertWrapper(len(self.label_indices)).to(device)
                self.bert.load_state_dict(torch.load(f))


    def _populate_labels(self, corpus):
        self.label_indices = {None: 0}
        self.label_indices_t = {0: None}
        for document in corpus.documents:
            for label in document.labels:
                if label not in self.label_indices:
                    index = len(self.label_indices)
                    self.label_indices[label] = index
                    self.label_indices_t[index] = label

    def _initialize(self, corpus):
        if self.initialized:
            return
        self.log('Initializing model...')
        self._populate_labels(corpus)
        if self.use_crf:
            self.bert = BertCRF(len(self.label_indices)).to(device)
        else:
            self.bert = BertWrapper(len(self.label_indices)).to(device)
        self.optimizer = optim.Adam(self.bert.parameters(), lr=2e-5)

        self.initialized = True

    def _embed_batch(self, documents):
        max_length = 0
        for document in documents:
            max_length = max(max_length, 2 + len(document.tokens))
        batch_size = len(documents)

        embedded = torch.zeros(batch_size, max_length, dtype=torch.int64)
        labels = torch.zeros_like(embedded)

        for i, document in enumerate(documents):
            tokens_text = [t.text for t in document.tokens]
            # this function shows dumb warnings -- avoid them :)
            sys.stderr = open(os.devnull, "w")
            encoded = self.tokenizer.encode(tokens_text)
            sys.stderr = sys.__stderr__
            assert len(encoded) <= max_length
            for j in range(len(encoded)):
                embedded[i,j] = encoded[j]
                if 0 <= j-1 < len(document.labels):
                    label = document.labels[j-1]
                    labels[i,j] = self.label_indices.get(label, 0)
        return (embedded.to(device), labels.to(device))


    def train_epoch(self, train_corpus):
        self.bert.train()
        for i, batch in enumerate(train_corpus.iter_batches(self.batch_size)):
            inputs, labels = self._embed_batch(batch)
            print('%.2f%%            ' % (100 * (i*self.batch_size / len(train_corpus.documents))), end='\r')
            self.optimizer.zero_grad()
            loss = self.bert.loss(inputs, labels)
            loss.backward()
            self.optimizer.step()

    def predict(self, corpus, retokenize=True):
        if retokenize:
            corpus.retokenize(self.tokenizer.tokenize)
        with torch.no_grad():
            self.bert.eval()
            for batch in corpus.iter_batches(self.batch_size, shuffle=False):
                inputs, labels = self._embed_batch(batch)
                predictions = self.bert(inputs)
                # logits: float32[batch_size, max_length, n_classes]
                for i, document in enumerate(batch):
                    document_length = len(document.tokens)
                    document.predictions = [self.label_indices_t[int(idx)] for idx in predictions[i,1:1+document_length]]
            
    def train(self, train_corpus, dev_corpus, gamma=0.75):
        train_corpus = deepcopy(train_corpus)
        dev_corpus = deepcopy(dev_corpus)
        train_corpus.retokenize(self.tokenizer.tokenize)
        dev_corpus.retokenize(self.tokenizer.tokenize)
        self._initialize(train_corpus)
        if self.use_crf:
            self._pretrain_crf(train_corpus, dev_corpus)
        best_f1 = -1
        average_f1 = -1
        epoch = 0
        self.log('Starting training')
        while True:
            self.predict(dev_corpus, retokenize=False)
            result = evaluation.eval_bios_exact(dev_corpus)
            f1 = result.micro_f1
            self.log('Epoch %d' % epoch)
            self.log('Running average: %f' % (100*average_f1))
            self.log(result.table())
            if f1 > best_f1:
                best_f1 = f1
                self.log('Saving model')
                self.save()
            elif f1 <= average_f1:
                break
            average_f1 = gamma*average_f1 + (1-gamma)*f1
            self.train_epoch(train_corpus)
            epoch += 1
        self.load()


    def _pretrain_crf(self, train_corpus, dev_corpus, gamma=0.75):
        self.optimizer = optim.Adam(self.bert.crf.parameters())
        best_f1 = -1
        average_f1 = -1
        epoch = 0
        self.log('Pretraining CRF layer')
        while True:
            self.predict(dev_corpus, retokenize=False)
            result = evaluation.eval_bios_exact(dev_corpus)
            f1 = result.micro_f1
            self.log('Epoch %d' % epoch)
            self.log('Running average: %f' % (100*average_f1))
            self.log(result.table())
            if f1 > best_f1:
                best_f1 = f1
                self.log('Saving model')
                self.save()
            elif f1 <= average_f1:
                break
            average_f1 = gamma*average_f1 + (1-gamma)*f1
            self.train_epoch(train_corpus)
            epoch += 1
        self.load()
        self.optimizer = optim.Adam(self.bert.parameters(), lr=2e-5)
