import os
import errno
import random
import unicodedata
from copy import deepcopy

class Token:
    def __init__(self, text, features, includes_space=False):
        self.text = text
        self.features = list(features)
        self.includes_space = includes_space

    def __repr__(self):
        return '<<' + self.text + '>>'

class Document:
    def __init__(self, tokens, labels=None):
        self.tokens = list(tokens)
        if labels is not None and len(labels) != len(tokens):
            raise ValueError('labels must have the same length as tokens')
        self.labels = list(labels)
        self.predictions = None
        self.missed_spans = []

    def __repr__(self):
        return '[[' + ' '.join([token.text for token in self.tokens]) + ']]'

    def retokenize(self, tokenizer, bios=True):
        # assume 1 or more new tokens for each old token
        new_tokens = []
        new_labels = []
        for old_token, old_label in zip(self.tokens, self.labels):
            new_subtokens = tokenizer(old_token.text)
            first_subtoken = True
            for new_subtoken in new_subtokens:
                if first_subtoken:
                    new_features = old_token.features + [('token start', True)]
                else:
                    new_features = [('token start', False)]
                new_tokens.append(Token(new_subtoken, new_features))
                new_labels.append(old_label)
                if bios and old_label.startswith('B'):
                    old_label = 'I' + old_label[1:]
                first_subtoken = False

        self.tokens = new_tokens
        self.labels = new_labels


class Corpus:
    def __init__(self, documents):
        self.documents = documents

    def retokenize(self, tokenizer, bios=True):
        for document in self.documents:
            document.retokenize(tokenizer, bios)

    def split_documents(self, max_length=8192):
        new_documents = []
        for document in self.documents:
            if len(document.tokens) > max_length:
                for start_idx in range(0, len(document.tokens), max_length):
                    new_tokens = document.tokens[start_idx:start_idx+max_length]
                    if document.labels is not None:
                        new_labels = document.labels[start_idx:start_idx+max_length]
                    else:
                        new_labels = None
                    new_documents.append(Document(new_tokens, new_labels))
            else:
                new_documents.append(document)
        self.documents = new_documents


    def iter_batches(self, batch_size, shuffle=True):
        if shuffle:
            documents = self.documents[:]
            random.shuffle(documents)
        else:
            documents = self.documents
        for i in range(0, len(documents), batch_size):
            yield documents[i:i+batch_size]

    def split_corpus(self, p_dev):
        docs_1 = []
        docs_2 = []
        for document in self.documents:
            if random.random() > p_dev:
                docs_1.append(deepcopy(document))
            else:
                docs_2.append(deepcopy(document))
        return Corpus(docs_1), Corpus(docs_2)
