import torch
import torch.nn as nn


# close enough :P
inf = 1e4

START = -2
END = -1

# returns log(exp(a) @ exp(b)), in a numerically stable manner
def log_matmul(a, b, epsilon=1e-200):
    max_a = torch.max(a, dim=-1)[0].view(a.shape[:-1] +  (1,))
    max_b = torch.max(b, dim=0)[0].view((1,) + b.shape[1:])
    a_exp = torch.exp(a - max_a)
    b_exp = torch.exp(b - max_b)
    prod = a_exp @ b_exp
    return (torch.log(prod) + max_a + max_b).view(prod.shape)

class CRFOutputLayer(nn.Module):
    def __init__(self, n_units, n_labels):
        super().__init__()
        self.n_units = n_units
        self.n_tags = n_labels + 2 # for start and end state
        self.in2feats = nn.Linear(n_units, self.n_tags)
        self.transitions = nn.Parameter(0.1 * torch.randn(self.n_tags, self.n_tags))
        # never transition to START or from END
        self.transitions.data[:, START] = -inf
        self.transitions.data[END, :] = -inf

    def _Q(self, feats):
        # feats: float[batch_size, sequence_length, self.n_tags]
        batch_size = feats.shape[0]
        sequence_length = feats.shape[1]
        logits = feats.new_full((batch_size, self.n_tags), -inf)
        logits[:, START] = 0
        # logits: float[batch_size, self.n_tags]
        for i in range(sequence_length):
            fi = feats[:,i,:]
            # fi: float[batch_size, self.n_tags]
            logits = log_matmul(logits, self.transitions) + fi
        logits = log_matmul(logits, self.transitions)
        return logits[:,END]
        #return: float[batch_size]
    
    def _score(self, feats, tags):
        # feats: float[batch_size, sequence_length, n_tags]
        # tags: long[batch_size, sequence_length]
        batch_size, sequence_length, _ = feats.shape
        s = feats.new_zeros(batch_size)
        # s: float[batch_size]
        last_tag = tags.new_full((batch_size,), START)
        # last_tag: long[batch_size]
        for i in range(sequence_length):
            fi = feats[:,i,:]
            # fi: float[batch_size, n_tags]
            ti = tags[:,i]
            # ti: long[batch_size]

            # get the transition scores for each element of the batch
            indices = last_tag * self.n_tags + ti
            # indices: long[batch_size]
            transition_scores = torch.take(self.transitions, indices)
            
            # get feature scores for each element of the batch
            feature_scores = torch.gather(fi, 1, ti.view(batch_size, 1)).view(batch_size)
            # feature_scores: float[batch_size]

            s += transition_scores + feature_scores
            last_tag = ti

        # score the final transition to END
        indices = last_tag * self.n_tags + tags.new_full((batch_size,), END)
        transition_scores = torch.take(self.transitions, indices)
        s += transition_scores

        return s
            
    def forward(self, x):
        # x : float[batch_size, sequence_length, n_units]
        feats = self.in2feats(x)
        # feats: float[batch_size, sequence_length, n_tags]
        batch_size, sequence_length, _ = feats.shape
        state_scores = x.new_full((batch_size, sequence_length+1, self.n_tags), -inf)
        # state_scores: float[batch_size, sequence_length+1, self.n_tags]
        state_scores[:,0,START] = 0
        back_pointers = []
        for i in range(sequence_length):
            fi = feats[:,i,:]
            # fi: float[batch_size, n_tags]
            best_edge_scores = []
            # best_edge_scores: [float:[batch_size]]
            back_pointers_i = []
            for next_tag in range(self.n_tags):
                edge_scores = state_scores[:,i,:] + self.transitions[:, next_tag]
                # edge_scores: float[batch_size, self.n_tags]
                prev_tag = torch.argmax(edge_scores, dim=1)
                # prev_tag: long[batch_size]
                back_pointers_i.append(prev_tag)
                best_edge_scores.append(
                    torch.gather(edge_scores, 1, prev_tag.view(batch_size, 1)).view(batch_size)
                )
            best_edge_scores = torch.stack(best_edge_scores, dim=1)
            # best_edge_scores: float[batch_size, n_tags]
            new_state_scores = best_edge_scores + fi
            # new_state_scores: float[batch_size, n_tags
            back_pointers.append(back_pointers_i)
            state_scores[:,i+1,:] = new_state_scores
            #state_scores = torch.cat((state_scores, new_state_scores.view(1, -1)))

        # now handle the transition to END
        edge_scores = state_scores[:,-2,:] + self.transitions[:, END]
        # edge_scores: float[batch_size, self.n_tags]
        prev_tag = torch.argmax(edge_scores, dim=1)
        # prev_tag: long[batch_size]
        # this is our final path score
        score = torch.gather(edge_scores, 1, prev_tag.view(batch_size, 1)).view(batch_size)
        path = [prev_tag]
        for back_pointers_i in reversed(back_pointers[1:]):
            bp = path[-1].new_zeros(batch_size)
            for j, sj in enumerate(path[-1]):
                bp[j] = back_pointers_i[sj][j]
            path.append(bp)
            #path.append(back_pointers_i[path[-1]])
        path.reverse()
        return torch.stack(path, dim=1)

    def loss(self, x, y, lengths):
        # TODO: incorporate lengths
        # Probably not necessary, and lots of extra code
        feats = self.in2feats(x)
        nll = self._Q(feats) - self._score(feats, y)
        return torch.mean(nll)
