from collections import defaultdict
from functools import total_ordering

@total_ordering
class ClassResults:
    def __init__(self):
        self.tp = 0
        self.fp = 0
        self.fn = 0

    @property
    def precision(self):
        if self.tp == 0:
            return 0
        return self.tp / (self.tp + self.fp)

    @property
    def recall(self):
        if self.tp == 0:
            return 0
        return self.tp / (self.tp + self.fn)

    def fscore(self, beta=1):
        if self.precision == 0 or self.recall == 0:
            return 0
        return (1 + beta**2) * self.precision * self.recall / (beta**2 * self.precision + self.recall)

    @property
    def f1(self):
        return self.fscore(1)

    def __or__(self, other):
        r = ClassResults()
        r.tp = self.tp + other.tp
        r.fp = self.fp + other.fp
        r.fn = self.fn + other.fn
        return r

    def __eq__(self, other):
        return self.f1 == other.f1

    def __lt__(self, other):
        return self.f1 < other.f1



@total_ordering
class Results:
    def __init__(self):
        self.class_results = defaultdict(ClassResults)

    @property
    def tp(self):
        return sum(cr.tp for cr in self.class_results.values())

    @property
    def fp(self):
        return sum(cr.fp for cr in self.class_results.values())

    @property
    def fn(self):
        return sum(cr.fn for cr in self.class_results.values())

    @property
    def micro_precision(self):
        if self.tp == 0:
            return 0
        return self.tp / (self.tp + self.fp)

    @property
    def micro_recall(self):
        if self.tp == 0:
            return 0
        return self.tp / (self.tp + self.fn)

    def micro_fscore(self, beta=1):
        if self.micro_precision == 0 or self.micro_recall == 0:
            return 0
        return (1 + beta**2) * self.micro_precision * self.micro_recall / (beta**2 * self.micro_precision + self.micro_recall)

    @property
    def micro_f1(self):
        return self.micro_fscore(1)

    def __eq__(self, other):
        return self.micro_f1 == other.micro_f1

    def __lt__(self, other):
        return self.micro_f1 < other.micro_f1

    def __or__(self, other):
        r = Results()
        for classname in self.class_results.keys() | other.class_results.keys():
            r.class_results[classname] = self.class_results[classname] | other.class_results[classname]
        return r

    def table(self):
        rows = []
        rows.append(['class', 'tp', 'fp', 'fn', 'P', 'R', 'F'])
        rows.append('-')
        for cls in sorted(self.class_results.keys()):
            results = self.class_results[cls]
            rows.append([cls, results.tp, results.fp, results.fn, 100*results.precision, 100*results.recall, 100*results.f1])
        rows.append('-')
        rows.append(['Total', self.tp, self.fp, self.fn, 100*self.micro_precision, 100*self.micro_recall, 100*self.micro_f1])
        return table_str(rows)


def table_str(table, lpad=1, rpad=1):
    for row in table:
        if not isinstance(row, list):
            continue
        for i in range(len(row)):
            if isinstance(row[i], float):
                row[i] = '%.2f' % row[i]
    column_widths = []
    for row in table:
        if not isinstance(row, list):
            continue
        for i, cell in enumerate(row):
            if i >= len(column_widths):
                column_widths.append(0)
            column_widths[i] = max(column_widths[i], len(str(cell)) + lpad + rpad)
    hline = '+'
    for w in column_widths:
        hline += '-' * w
        hline += '+'
    
    lines = [hline]
    for row in table:
        line = '|'
        if isinstance(row, list):
            for i, cell in enumerate(row):
                cell = ' ' * lpad + str(cell) + ' ' * rpad
                while len(cell) < column_widths[i]:
                    cell += ' '
                line += cell + '|'
            lines.append(line)
        elif row == '-':
            lines.append(hline)
        else:
            lines.append(str(row))
    lines.append(hline)
    return '\n'.join(lines)
            

def spanset(labels):
    spans = set()
    current_start = None
    current_type = None
    for i, li in enumerate(labels):
        if li is None:
            li = 'O'
        if current_start is not None and not li.startswith('I'):
            spans.add((current_type, current_start, i))
            current_type = None
            current_start = None
        if isinstance(li, str) and li.startswith('B-'):
            current_start = i
            current_type = li[2:]
    if current_start:
        spans.add((current_type, current_start, len(labels)))
    return spans
    
def eval_bios_exact(corpus):
    res = Results()
    for document in corpus.documents:
        hypo_spans = spanset(document.predictions)
        gold_spans = spanset(document.labels)
        for hypo_span in hypo_spans:
            cls = hypo_span[0]
            class_results = res.class_results[cls]
            if hypo_span in gold_spans:
                class_results.tp += 1
            else:
                class_results.fp += 1
        for gold_span in gold_spans:
            cls = gold_span[0]
            class_results = res.class_results[cls]
            if gold_span not in hypo_spans:
                class_results.fn += 1
        for cls in document.missed_spans:
            # if some spans cound not be represented, i.e. because of tokenization
            class_results = res.class_results[cls]
            class_results.fn += 1
    return res
    
