"""
Usage:
  main.py train    <experiment-name> <architecture> [--restricted] (--parc-corpus=<corpus-path> | --ontonotes-corpus=<corpus-path> | --riqua-corpus=<corpus-path> | --conll2k-corpus=<corpus-path> | --chemdner-corpus=<corpus-path>) [--embeddings=<embeddings-file>]
  main.py evaluate <experiment-name> <architecture> [--restricted] (--parc-corpus=<corpus-path> | --ontonotes-corpus=<corpus-path> | --riqua-corpus=<corpus-path> | --conll2k-corpus=<corpus-path> | --chemdner-corpus=<corpus-path>)
  main.py predict  <experiment-name> <architecture> [--restricted] (--parc-corpus=<corpus-path> | --ontonotes-corpus=<corpus-path> | --riqua-corpus=<corpus-path> | --conll2k-corpus=<corpus-path> | --chemdner-corpus=<corpus-path>) -o <output-file>

Options:
  -e --embeddings=<embeddings-file>   The path to a pretrained embeddings file.  Should be in glove format
  --restricted                        Don't use handcrafted features
"""
from docopt import docopt
if __name__ == '__main__':
    arguments = docopt(__doc__)


import random
import os
from copy import deepcopy
from datetime import datetime

import torch
import torch.optim as optim

#from corpus import load_corpus, load_labels, iter_batches, load_parc_corpus_and_labels
import parc_loader
import ontonotes_loader
import riqua_loader
import conll2k_loader
import chemdner_loader

from sequence_labeler import SequenceLabeler
from models import SequenceLabelingModel
from bert_labeler import BertSequenceLabelingModel
from real_big_model import RealBigModel
import evaluation

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    
def log(*args):
    timestamp = datetime.now().replace(microsecond=0).isoformat(' ') + '| '
    filler = '\n' + ' ' * (len(timestamp) - 2) + '| '
    new_args = []
    for arg in args:
        arg = str(arg)
        arg = arg.replace('\n', filler)
        arg = timestamp + arg
        new_args.append(arg)

    if len(new_args) == 0:
        new_args = (filler[1:],)

    with open('logs/%s.txt' % EXPERIMENT_NAME, 'a') as f:
        print(*new_args, file=f)
    print(*new_args)

    
if __name__ == '__main__':
    EXPERIMENT_NAME = arguments['<experiment-name>']
    architecture = arguments['<architecture>']
    batch_size = 5
    cross_validate = False
    split_train = False
    restricted_features = arguments['--restricted']
    if arguments['--parc-corpus']:
        corpus_dir = arguments['--parc-corpus']
        load_corpus = parc_loader.load_corpus
        train_path = os.path.join(corpus_dir, 'train/')
        dev_path = os.path.join(corpus_dir, 'dev/')
        test_path = os.path.join(corpus_dir, 'test/')

    elif arguments['--ontonotes-corpus']:
        corpus_dir = arguments['--ontonotes-corpus']
        load_corpus = ontonotes_loader.load_corpus
        train_path = os.path.join(corpus_dir, 'train/')
        dev_path = os.path.join(corpus_dir, 'development/')
        test_path = os.path.join(corpus_dir, 'test/')

    elif arguments['--riqua-corpus']:
        cross_validate = True
        split_train = True
        corpus_path = arguments['--riqua-corpus']
        iter_xvalidation_splits = riqua_loader.iter_xvalidation_splits
        batch_size = 1
        #train_path = os.path.join(corpus_dir, 'train/')
        #dev_path = os.path.join(corpus_dir, 'dev/')
        #test_path = os.path.join(corpus_dir, 'test/')

    elif arguments['--conll2k-corpus']:
        corpus_dir = arguments['--conll2k-corpus']
        load_corpus = conll2k_loader.load_corpus
        train_path = os.path.join(corpus_dir, 'train.txt')
        dev_path = None
        test_path = os.path.join(corpus_dir, 'test.txt')

    elif arguments['--chemdner-corpus']:
        corpus_dir = arguments['--chemdner-corpus']
        load_corpus = chemdner_loader.load_corpus
        train_path = os.path.join(corpus_dir, 'training')
        dev_path = os.path.join(corpus_dir, 'development')
        test_path = os.path.join(corpus_dir, 'evaluation')

    
    if arguments['train']:
        embeddings_path = arguments['--embeddings']

        if not cross_validate:
            log('Loading train data...')
            train_corpus = load_corpus(train_path, restricted_features=restricted_features)

            if dev_path is None:
                train_corpus, dev_corpus = train_corpus.split_corpus(0.1)
            else:
                log('Loading dev data...')
                dev_corpus = load_corpus(dev_path, restricted_features=restricted_features)

            log('Loading test data...')
            test_corpus = load_corpus(test_path, restricted_features=restricted_features)

            split_iter = [(train_corpus, dev_corpus, test_corpus)]

        else:
            split_iter = iter_xvalidation_splits(corpus_path, log, restricted_features=restricted_features)

        result = evaluation.Results()
        for train_corpus, dev_corpus, test_corpus in split_iter:
            if split_train:
                train_corpus.split_documents()
                
            if architecture in {'baseline', 'lstm', 'crf', 'lstm-crf'}:
                model = SequenceLabelingModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, embeddings_path=embeddings_path, log=log)
            elif architecture in {'bert', 'bert-crf'}:
                model = BertSequenceLabelingModel(EXPERIMENT_NAME, architecture, batch_size=batch_size, log=log)
            else:
                model = RealBigModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, log=log)

            model.train(train_corpus, dev_corpus)
            model.predict(test_corpus)
            split_result = evaluation.eval_bios_exact(test_corpus)
            if cross_validate:
                log('SPLIT RESULTS:\n')
                log(split_result.table())
            result |= split_result
        log('FINAL RESULTS:\n')
        log(result.table())

    elif arguments['evaluate']:
        if architecture in {'baseline', 'lstm', 'crf', 'lstm-crf'}:
            model = SequenceLabelingModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, embeddings_path=embeddings_path, log=log)
        elif architecture in {'bert', 'bert-crf'}:
            model = BertSequenceLabelingModel(EXPERIMENT_NAME, architecture, batch_size=batch_size, log=log)
        else:
            model = RealBigModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, log=log)

        model.load()
        test_corpus = load_corpus(test_path, restricted_features=restricted_features)
        model.predict(test_corpus)
        result = evaluation.eval_bios_exact(test_corpus)
        print(result.table())

    elif arguments['predict']:
        with open(arguments['<output-file>'], 'w') as f:
            if architecture in {'baseline', 'lstm', 'crf', 'lstm-crf'}:
                model = SequenceLabelingModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, embeddings_path=embeddings_path, log=log)
            elif architecture in {'bert', 'bert-crf'}:
                model = BertSequenceLabelingModel(EXPERIMENT_NAME, architecture, batch_size=batch_size, log=log)
            else:
                model = RealBigModel(EXPERIMENT_NAME, 300, architecture, batch_size=batch_size, log=log)

            model.load()
            if dev_path is not None:
                dev_corpus = load_corpus(dev_path, restricted_features=restricted_features)
            else:
                dev_corpus = load_corpus(test_path, restricted_features=restricted_features)
            model.predict(dev_corpus)
            for document in dev_corpus.documents:
                print('TOKEN           LABEL           PREDICTION', file=f)
                for token, label, prediction in zip(document.tokens, document.labels, document.predictions):
                    line = token.text
                    while len(line) < 16:
                        line += ' '
                    line += label
                    while len(line) < 32:
                        line += ' '
                    line += prediction
                    print(line, file=f)
                print(file=f)
