import os
import pickle
import random

import torch
import torch.optim as optim

from sequence_labeler import SequenceLabeler
import evaluation

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class SequenceLabelingModel:
    def __init__(self, model_name, n_units, architecture, batch_size=5, embeddings_path=None, log=print):
        self.model_name = model_name
        self.feature_indices = None
        self.feature_indices_t = None
        self.label_indices = None
        self.label_indices_t = None
        self.n_units = n_units
        self.architecture = architecture
        self.batch_size = batch_size
        self.embeddings_path = embeddings_path
        self.sequence_labeler = None
        self.initialized = False
        self.log=log

    def _populate_features(self, corpus, min_count=5):
        feature_counts = {}
        for document in corpus.documents:
            for token in document.tokens:
                for feature in token.features:
                    if feature not in feature_counts:
                        feature_counts[feature] = 0
                    feature_counts[feature] += 1
        self.feature_indices = {}
        self.feature_indices_t = {}
        for feature, count in feature_counts.items():
            if count >= min_count:
                index = len(self.feature_indices)
                self.feature_indices[feature] = index
                self.feature_indices_t[index] = feature

    def _populate_labels(self, corpus):
        self.label_indices = {None: 0}
        self.label_indices_t = {0: None}
        for document in corpus.documents:
            for label in document.labels:
                if label not in self.label_indices:
                    index = len(self.label_indices)
                    self.label_indices[label] = index
                    self.label_indices_t[index] = label

    def _load_embeddings(self):
        if self.embeddings_path is None:
            return
        self.log('Loading embeddings...')
        loaded_words = set()
        with open(self.embeddings_path, 'r') as f:
            for line in f:
                line = line.split()
                word = line[0].lower()
                if word in loaded_words:
                    continue
                word_index = None
                if word in self.feature_indices:
                    word_index = self.feature_indices[word]
                elif ('word', word) in self.feature_indices:
                    word_index = self.feature_indices[('word', word)]
                if word_index is not None:
                    dim = min(len(line[1:]), self.sequence_labeler.n_units)
                    word_vector = torch.tensor([float(xi) for xi in line[1:]], dtype=torch.float32).to(device)
                    self.sequence_labeler.embedding.weight.data[word_index,:dim] = word_vector[:dim]
                    loaded_words.add(word)


    def _initialize(self, corpus):
        if self.initialized:
            return
        self.log('Initializing model...')
        self._populate_features(corpus)
        self._populate_labels(corpus)

        self.sequence_labeler = SequenceLabeler(self.architecture, len(self.feature_indices), self.n_units, len(self.label_indices)).to(device)
        self.optimizer = optim.Adam(self.sequence_labeler.parameters())
        self._load_embeddings()
        self.initialized = True

    def _embed_batch(self, documents, dropout_rate=0):
        max_length = 0
        for document in documents:
            max_length = max(max_length, len(document.tokens))
        batch_size = len(documents)
        
        offsets = torch.zeros(batch_size, max_length, dtype=torch.int64)
        feature_indices = []

        lengths = torch.tensor([len(document.tokens) for document in documents])

        labels = torch.zeros(batch_size, max_length, dtype=torch.int64)
        
        for i, document in enumerate(documents):
            for j in range(max_length):
                offsets[i,j] = len(feature_indices)
                if j < len(document.tokens):
                    token = document.tokens[j]
                    for feature in token.features:
                        if feature in self.feature_indices and random.random() >= dropout_rate:
                            feature_indices.append(self.feature_indices[feature])
            for j, label in enumerate(document.labels):
                labels[i,j] = self.label_indices.get(label, 0)

        indices = torch.tensor(feature_indices, dtype=torch.int64)

        return (indices.to(device), offsets.to(device)), lengths.to(device), labels.to(device)


    def save(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        d = dict(self.__dict__)
        d['sequence_labeler'] = None
        del d['log']
        del d['model_name']
        with open(path, 'wb') as f:
            pickle.dump(d, f)
            if self.initialized:
                torch.save(self.sequence_labeler.state_dict(), f)

    def load(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        with open(path, 'rb') as f:
            d = pickle.load(f)
            for k, v in d.items():
                self.__dict__[k] = v
            if self.initialized:
                self.sequence_labeler = SequenceLabeler(self.architecture, len(self.feature_indices), self.n_units, len(self.label_indices)).to(device)
                self.sequence_labeler.load_state_dict(torch.load(f))
        
    def train_epoch(self, train_corpus):
        self.sequence_labeler.train()
        for i, batch in enumerate(train_corpus.iter_batches(self.batch_size)):
            (indices, offsets), lengths, labels = self._embed_batch(batch, 0.5)
            print('%.2f%%            ' % (100 * (i*self.batch_size / len(train_corpus.documents))), end='\r')
            self.optimizer.zero_grad()
            loss = self.sequence_labeler.loss((indices, offsets), labels, lengths)
            loss.backward()
            self.optimizer.step()

    def train(self, train_corpus, dev_corpus, gamma=0.75):
        self._initialize(train_corpus)
        best_f1 = -1
        average_f1 = -1
        epoch = 0
        self.log('Starting training')
        while True:
            self.predict(dev_corpus)
            result = evaluation.eval_bios_exact(dev_corpus)
            f1 = result.micro_f1
            self.log('Epoch %d' % epoch)
            self.log('Running average: %f' % (100*average_f1))
            self.log(result.table())
            if f1 > best_f1:
                best_f1 = f1
                self.log('Saving model')
                self.save()
            elif f1 <= average_f1:
                break
            average_f1 = gamma*average_f1 + (1-gamma)*f1
            self.train_epoch(train_corpus)
            epoch += 1
        self.load()

    def predict(self, corpus):
        self.sequence_labeler.eval()
        for batch in corpus.iter_batches(self.batch_size, shuffle=False):
            (indices, offsets), lengths, labels = self._embed_batch(batch, 0)
            predictions = self.sequence_labeler((indices, offsets))
            # predictions: int64[batch_size, max_length]
            for i, document in enumerate(batch):
                document_length = len(document.tokens)
                document.predictions = [self.label_indices_t[int(idx)] for idx in predictions[i,:document_length]]
