import sys
import os
import re
import xml.etree.ElementTree as ET

from corpus import Token, Document, Corpus

'''
Features implemented:
* normalized text
* lemma
* pos
* start and end of constituents
'''

def leaf_words(node):
    if node.tag == 'WORD':
        return [node]
    leaves = []
    for child in node:
        leaves += leaf_words(child)
    return leaves

def words_spans(words, preserve_nested=False, span_types=None):
    open_spans = {}
    closed_spans = []
    for i, word in enumerate(words):
        current_spans = set()
        for attribution in word:
            for attributionRole in attribution:
                role = attributionRole.attrib['roleValue']
                if span_types is None or role in span_types:
                    span_id = (role, attribution.attrib['id'])
                    current_spans.add(span_id)
        # close any open spans that don't continue here
        still_open_spans = {}
        for span_id in open_spans:
            if span_id in current_spans:
                still_open_spans[span_id] = open_spans[span_id]
            else:
                role, start = open_spans[span_id]
                closed_spans.append((role, start, i))
        open_spans = still_open_spans
        # open any new spans
        for span_id in current_spans:
            if span_id not in open_spans:
                role = span_id[0]
                open_spans[span_id] = (role, i)
    # close any spans still open
    for span_id in open_spans:
        role, start = open_spans[span_id]
        closed_spans.append((role, start, i))

    # for some reason there were some duplicates :/
    closed_spans = list(set(closed_spans))

    # filter nested spans
    if preserve_nested:
        return closed_spans
    else:
        filtered = []
        for i, (inner_role, inner_l, inner_r) in enumerate(closed_spans):
            for outer_role, outer_l, outer_r in closed_spans[:i] + closed_spans[i+1:]:
                if (inner_l, inner_r) == (outer_l, outer_r):
                    # if the spans exactly coincide (which does happen sometimes) only take the content
                    if inner_role != 'content':
                        break
                # completely inside
                if inner_l > outer_l and inner_r < outer_r:
                    break
                # shared l
                if inner_l == outer_l and inner_r < outer_r:
                    break
                # shared r
                if inner_r == outer_r and inner_l > outer_l:
                    break
            else:
                filtered.append((inner_role, inner_l, inner_r))
        return filtered


def document_labels(path, span_types=None):
    tree = ET.parse(path)
    root = tree.getroot()
    words = leaf_words(root)
    spans = words_spans(words, span_types=span_types)
    labels = []
    for i in range(len(words)):
        labels.append('O')
    for span_type, l, r in spans:
        labels[l] = 'B-' + span_type
        for i in range(l+1, r):
            labels[i] = 'I-' + span_type
    return labels


def document_tokens(path, restricted_features):
    tree = ET.parse(path)
    root = tree.getroot()
    tokens = node_tokens(root, restricted_features)
    return tokens

def load_document(path, span_types, restricted_features):
    tokens = document_tokens(path, restricted_features)
    labels = document_labels(path, span_types=span_types)
    return Document(tokens, labels)

def node_tokens(node, restricted_features, const=[], const_start = [], const_end = []):
    if restricted_features:
        features_to_use = {'word'}
    else:
        features_to_use = {
            'word',
            'constituent',
            'lemma',
            'pos',
        }
    if node.tag == 'WORD':
        wf = word_features(node, features_to_use)
        if 'constituent' in features_to_use:
            for constituent in const_start:
                wf.append(('constituent start', constituent))
            for constituent in const_end:
                wf.append(('constituent end', constituent))
            for constituent in const:
                wf.append(('constituent', constituent))
        # cue features
        '''
        is_cue = False
        for attribution in node:
            for attributionRole in attribution:
                if attributionRole.attrib['roleValue'] == 'cue':
                    is_cue = True
                    break
        wf.append(('cue', is_cue))       
        '''
        return [Token(node.attrib['text'], wf)]

    child_tokens = []
    for i, child in enumerate(node):
        if i == 0:
            child_const_start = const_start + [node.tag]
        else:
            child_const_start = []
        if i == len(node) - 1:
            child_const_end = const_end + [node.tag]
        else:
            child_const_end = []
        child_tokens += node_tokens(child, restricted_features, const + [node.tag], child_const_start, child_const_end)
    return child_tokens
 
def word_features(node, features_to_use):
    wf = []
    if 'word' in features_to_use:
        wf.append(("word", normalize_text(node.attrib['text'])))
    if 'lemma' in features_to_use:
        wf.append(("lemma", normalize_text(node.attrib['lemma'])))
    if 'pos' in features_to_use:
        wf.append(("pos", node.attrib['pos']))
    return wf

dig = re.compile('[0-9]')
def normalize_text(s):
    #such a clean and beautiful interface...
    if s is None:
        return s
    s = s.lower()
    s = re.sub(dig, '0', s)
    s = s.replace("``", '"')
    s = s.replace("''", '"')
    return s


def load_corpus(path, span_types=None, restricted_features=False):
    documents = []
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith('.xml'):
                fpath = os.path.join(dirpath, filename)
                documents.append(load_document(fpath, span_types=span_types, restricted_features=restricted_features))
    assert len(documents) > 0
    return Corpus(documents)

if __name__ == '__main__':
    import sys
    corpus = load_corpus(sys.argv[1], restricted_features=True)
    breakpoint()
