import random
from copy import deepcopy
import os
import sys
import pickle

import torch
import torch.nn as nn
import torch.optim as optim

from transformers import BertModel, BertForTokenClassification, BertTokenizer
from sequence_labeler import LSTMSequenceEncoder

from crf_layer import CRFOutputLayer
import evaluation

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class BertWrapper(nn.Module):
    def __init__(self):
        super().__init__()
        self.bert = BertModel.from_pretrained('bert-base-uncased')
        self.n_units = self.bert.config.hidden_size
        self.max_length = self.bert.config.max_position_embeddings

    def forward(self, inputs):
        sequence_length = inputs.shape[1]
        if sequence_length <= self.max_length:
            output = self.bert(inputs)[0]
        else:
            output_windows = []
            for window_start in range(0, sequence_length, self.max_length//2):
                input_window = inputs[:,window_start:window_start+self.max_length]
                window_length = input_window.shape[1]
                # input_window: float32[batch_size, window_length]
                output_window = self.bert(input_window)[0]
                # output_window: float32[batch_size, window_length]
                output_windows.append(output_window)
            window_start_index = self.max_length//4
            window_end_index = 3*self.max_length//4
            parts = []
            parts.append(output_windows[0][:,:window_end_index])
            for output_window in output_windows[1:-1]:
                parts.append(output_window[:,window_start_index:window_end_index])
            parts.append(output_windows[-1][:,window_start_index:])
            output = torch.cat(parts, dim=1)
        return output


class BertLSTMCRF(nn.Module):
    def __init__(self, vocab_size, n_units, n_labels):
        super().__init__()
        self.vocab_size = vocab_size
        self.n_units = n_units
        self.n_labels = n_labels
        self.bert = BertWrapper()
        self.bert_projection = nn.Linear(self.bert.n_units, n_units)
        self.embedding = nn.EmbeddingBag(self.vocab_size, self.n_units)
        self.lstm = LSTMSequenceEncoder(self.n_units)
        self.crf = CRFOutputLayer(self.n_units, self.n_labels)

    def forward(self, bert_tokens, indices, offsets):
        # bert_tokens: int64[batch_size, sequence_length]
        # indices: int64[n_features]
        # offsets: int64[batch_size, sequence_length]
        batch_size, sequence_length = bert_tokens.shape
        embedded = self.embedding(indices, offsets.view(-1)).view(batch_size, sequence_length, self.n_units)
        # embedded: float32[batch_size, sequence_length, self.n_units]
        bert_embedded = self.bert_projection(self.bert(bert_tokens))
        # bert_embedded: float32[batch_size, sequence_length, self.n_units]
        embedded = embedded + bert_embedded
    
        lstm_encoded = embedded + self.lstm(embedded)
        # lstm_encoded: float32[batch_size, sequence_length, self.n_units]
        out = self.crf(lstm_encoded)
        # out: int64[batch_size, sequence_length]
        return out

    def loss(self, bert_tokens, indices, offsets, labels):
        # bert_tokens: int64[batch_size, sequence_length]
        # indices: int64[n_features]
        # offsets: int64[batch_size, sequence_length]
        batch_size, sequence_length = bert_tokens.shape

        embedded = self.embedding(indices, offsets.reshape(-1)).view(batch_size, sequence_length, self.n_units)
        # embedded: float32[batch_size, sequence_length, self.n_units]

        # if the sequence is too long to backprop, only select a random window
        if sequence_length > self.bert.max_length:
            start_index = random.randrange(sequence_length - self.bert.max_length + 1)
            bert_tokens = bert_tokens[:,start_index:start_index + self.bert.max_length]
            embedded = embedded[:,start_index:start_index+self.bert.max_length]
            labels = labels[:,start_index:start_index+self.bert.max_length]
            sequence_length = self.bert.max_length

        bert_embedded = self.bert_projection(self.bert(bert_tokens))
        # bert_embedded: float32[batch_size, sequence_length, self.n_units]
        embedded = embedded + bert_embedded
    
        lstm_encoded = embedded + self.lstm(embedded)
        # lstm_encoded: float32[batch_size, sequence_length, self.n_units]
        return self.crf.loss(lstm_encoded, labels, None)

    def pretraining_parameters(self):
        return list(self.bert_projection.parameters()) + list(self.embedding.parameters()) + list(self.lstm.parameters()) + list(self.crf.parameters())

class RealBigModel:
    def __init__(self, model_name, n_units, architecture, batch_size, log=print):
        self.model_name = model_name
        self.n_units = n_units
        assert architecture == 'bert-lstm-crf'
        self.batch_size = batch_size
        self.log = log
        self.initialized = False
        self.label_indices = None
        self.label_indices_t = None
        self.bert_tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')

    def save(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        d = dict(self.__dict__)
        d['bfm'] = None
        del d['log']
        del d['model_name']
        with open(path, 'wb') as f:
            pickle.dump(d, f)
            if self.initialized:
                torch.save(self.bfm.state_dict(), f)

    def load(self, path=None):
        if path is None:
            path = os.path.join('models', self.model_name + '.model')
        with open(path, 'rb') as f:
            d = pickle.load(f)
            for k, v in d.items():
                self.__dict__[k] = v
            if self.initialized:
                self.bfm = BertLSTMCRF(len(self.feature_indices), self.n_units, len(self.label_indices)).to(device)
                self.bfm.load_state_dict(torch.load(f))


    def _populate_features(self, corpus, min_count=5):
        feature_counts = {}
        for document in corpus.documents:
            for token in document.tokens:
                for feature in token.features:
                    if feature not in feature_counts:
                        feature_counts[feature] = 0
                    feature_counts[feature] += 1
        self.feature_indices = {}
        self.feature_indices_t = {}
        for feature, count in feature_counts.items():
            if count >= min_count:
                index = len(self.feature_indices)
                self.feature_indices[feature] = index
                self.feature_indices_t[index] = feature

    def _populate_labels(self, corpus):
        self.label_indices = {None: 0}
        self.label_indices_t = {0: None}
        for document in corpus.documents:
            for label in document.labels:
                if label not in self.label_indices:
                    index = len(self.label_indices)
                    self.label_indices[label] = index
                    self.label_indices_t[index] = label


    def _initialize(self, corpus):
        if self.initialized:
            return
        self.log('Initializing model...')
        self._populate_features(corpus)
        self._populate_labels(corpus)
        self.bfm = BertLSTMCRF(len(self.feature_indices), self.n_units, len(self.label_indices)).to(device)
        self.optimizer = optim.Adam(self.bfm.parameters(), lr=2e-5)

        self.initialized = True

    def _embed_batch(self, documents, dropout_rate=0):
        max_length = 0
        for document in documents:
            max_length = max(max_length, 2+len(document.tokens))
        batch_size = len(documents)

        offsets = torch.zeros(batch_size, max_length, dtype=torch.int64)
        feature_indices = []

        bert_encoded = torch.zeros(batch_size, max_length, dtype=torch.int64)

        lengths = torch.tensor([len(document.tokens) for document in documents])

        labels = torch.zeros(batch_size, max_length, dtype=torch.int64)
        
        for i, document in enumerate(documents):
            # get bert_encoded
            tokens_text = [t.text for t in document.tokens]
            # this function shows dumb warnings -- avoid them :)
            sys.stderr = open(os.devnull, "w")
            encoded = self.bert_tokenizer.encode(tokens_text)
            sys.stderr = sys.__stderr__
            assert len(encoded) <= max_length
            for j in range(len(encoded)):
                bert_encoded[i,j] = encoded[j]
            
            # get feature indices and offsets
            for j in range(max_length):
                offsets[i,j] = len(feature_indices)
                if 0 <= j-1 < len(document.tokens):
                    token = document.tokens[j-1]
                    for feature in token.features:
                        if feature in self.feature_indices and random.random() >= dropout_rate:
                            feature_indices.append(self.feature_indices[feature])
            for j, label in enumerate(document.labels):
                labels[i,j+1] = self.label_indices.get(label, 0)

        indices = torch.tensor(feature_indices, dtype=torch.int64)

        return bert_encoded.to(device), indices.to(device), offsets.to(device), lengths.to(device), labels.to(device)

    def train_epoch(self, train_corpus):
        self.bfm.train()
        for i, batch in enumerate(train_corpus.iter_batches(self.batch_size)):
            bert_encoded, indices, offsets, lengths, labels = self._embed_batch(batch, 0.5)
            print('%.2f%%            ' % (100 * (i*self.batch_size / len(train_corpus.documents))), end='\r')
            self.optimizer.zero_grad()
            loss = self.bfm.loss(bert_encoded, indices, offsets, labels)
            loss.backward()
            self.optimizer.step()

    def predict(self, corpus, retokenize=True):
        if retokenize:
            corpus.retokenize(self.bert_tokenizer.tokenize)
        with torch.no_grad():
            self.bfm.eval()
            for batch in corpus.iter_batches(self.batch_size, shuffle=False):
                bert_encoded, indices, offsets, lengths, labels = self._embed_batch(batch, 0)
                predictions = self.bfm(bert_encoded, indices, offsets)
                for i, document in enumerate(batch):
                    document_length = len(document.tokens)
                    document.predictions = [self.label_indices_t[int(idx)] for idx in predictions[i,1:1+document_length]]
            
    def train(self, train_corpus, dev_corpus, gamma=0.75):
        train_corpus = deepcopy(train_corpus)
        dev_corpus = deepcopy(dev_corpus)
        train_corpus.retokenize(self.bert_tokenizer.tokenize)
        dev_corpus.retokenize(self.bert_tokenizer.tokenize)
        self._initialize(train_corpus)
        self._pretrain(train_corpus, dev_corpus)
        best_f1 = -1
        average_f1 = -1
        epoch = 0
        self.log('Starting training')
        while True:
            self.predict(dev_corpus, retokenize=False)
            result = evaluation.eval_bios_exact(dev_corpus)
            f1 = result.micro_f1
            self.log('Epoch %d' % epoch)
            self.log('Running average: %f' % (100*average_f1))
            self.log(result.table())
            if f1 > best_f1:
                best_f1 = f1
                self.log('Saving model')
                self.save()
            elif f1 <= average_f1:
                break
            average_f1 = gamma*average_f1 + (1-gamma)*f1
            self.train_epoch(train_corpus)
            epoch += 1
        self.load()


    def _pretrain(self, train_corpus, dev_corpus, gamma=0.75):
        self.optimizer = optim.Adam(self.bfm.pretraining_parameters())
        best_f1 = -1
        average_f1 = -1
        epoch = 0
        self.log('Pretraining Non-bert layers')
        while True:
            self.predict(dev_corpus, retokenize=False)
            result = evaluation.eval_bios_exact(dev_corpus)
            f1 = result.micro_f1
            self.log('Epoch %d' % epoch)
            self.log('Running average: %f' % (100*average_f1))
            self.log(result.table())
            if f1 > best_f1:
                best_f1 = f1
                self.log('Saving model')
                self.save()
            elif f1 <= average_f1:
                break
            average_f1 = gamma*average_f1 + (1-gamma)*f1
            self.train_epoch(train_corpus)
            epoch += 1
        self.load()
        self.optimizer = optim.Adam(self.bfm.parameters(), lr=2e-5)
