import numpy as np
from task_properties import get_task_properties
from fetch_f1s import get_f1s

from sklearn import linear_model, preprocessing
from sklearn.pipeline import make_pipeline

import statsmodels.api as sm

import pickle
from collections import defaultdict

from scipy.special import logit
def sigmoid(x):
    return np.exp(x) / (1 + np.exp(x))


def to_log(f1, a=0.2):
    f1 /= 100
    padded = (1-f1)*a + f1*(1-a)
    return logit(padded)

def from_log(lf1, a=0.2):
    padded = sigmoid(lf1)
    f1 = (padded - a) / (1-2*a)
    f1 = np.maximum(f1, 0)
    f1 = np.minimum(f1, 1)
    return 100*f1

def perform_regression(f1s, do_search=False):
    encoded_splits, dim_labels, _ = encode_splits(f1s)
    
    if do_search:
        log_alpha = 0
        logit_ratio = 0

        alpha = np.exp(log_alpha)
        ratio = sigmoid(logit_ratio)

        # coordinate descent on alpha and ratit
        for step in range(3):
            print("finding alpha")
            log_alpha, mse = coordinate_search(lambda la: cross_validate_regression(encoded_splits, np.exp(la), ratio)[0], log_alpha)
            alpha = np.exp(log_alpha)
            print("alpha =", alpha)
            print()
            print("finding ratio")
            logit_ratio, mse = coordinate_search(lambda lr: cross_validate_regression(encoded_splits, alpha, sigmoid(lr))[0], logit_ratio)
            ratio = sigmoid(logit_ratio)
            print("ratio =", ratio)
            print()
    else:
        alpha = 0
        ratio = 0.5

    mse, mae, coef, intercept = cross_validate_regression(encoded_splits, alpha, ratio)
    print("RMSE =", mse**0.5)
    print("MAE", mae)
    print()

    A, b, _ = encode(f1s, dim_labels)
    final_model = fit_model(A, to_log(b), alpha, ratio)


    
    fws = []
    for feature, weight in zip(dim_labels, final_model[1].coef_):
        fws.append((feature, weight))
    fws.sort(key=lambda fw: -abs(fw[1]))
    for f, w in fws:
        if isinstance(f, tuple):
            f = " * ".join(f)
        print(f + ":", w)
    print('intercept:', final_model[1].intercept_)

    
def coordinate_search(func, estimate=0, iterations=5):
    step = 1
    for iteration in range(iterations):
        estimate, minimum = linear_search(func, estimate, step)
        print(iteration, estimate, minimum**0.5)
        step /= 2
    return estimate, minimum
        

def linear_search(func, estimate=0, step=1):
    minimum = func(estimate)
    argmin = estimate
    value = func(estimate-step)
    if value < minimum:
        direction = -1
        estimate -= step
        minimum = value
        argmin = estimate
    else:
        value = func(estimate+step)
        if value < minimum:
            direction = +1
            estimate += step
            minimum = value
            argmin = estimate
        else:
            return estimate, minimum
    while True:
        estimate += (step*direction)
        value = func(estimate)
        if value < minimum:
            minimum = value
            argmin = estimate
        else:
            return argmin, minimum


def cross_validate_regression(splits, alpha, ratio):
    split_coefs = []
    split_intercepts = []
    sse = 0
    sae = 0
    for holdout_index in range(len(splits)):
        holdout_A = splits[holdout_index][0]
        holdout_b = splits[holdout_index][1]
        train_A = np.concatenate([splits[i][0] for i in range(len(splits)) if i != holdout_index])
        train_b = np.concatenate([splits[i][1] for i in range(len(splits)) if i != holdout_index])

        model = fit_model(train_A, to_log(train_b), alpha, ratio)
        split_coefs.append(model[1].coef_)
        split_intercepts.append(model[1].intercept_)
        square_error = np.mean((from_log(model.predict(holdout_A)) - holdout_b)**2)
        absolute_error = np.mean(abs(from_log(model.predict(holdout_A)) - holdout_b))
        sse += square_error
        sae += absolute_error
    mse = sse / len(splits)
    mae = sae / len(splits)
    return mse, mae, split_coefs, split_intercepts


def detailed_error_data(f1s):
    def model_from_features(feats, dim_labels):
        model_properties = []
        try:
            for prop in ['Feat', 'CRF', 'LSTM', 'BERT']:
                if feats[dim_labels.index(prop)]:
                    model_properties.append(prop)
        except ValueError:
            return ()
        return tuple(sorted(model_properties))

    # for each model, a list of (pred, gold) pairs
    model_pairs = defaultdict(list)
    
    encoded_splits, dim_labels, corpus_tasks = encode_splits(f1s)
    # we already know that we just use OLS, so no hyperparameter search
    for holdout_index in range(len(encoded_splits)):
        corpus, task = corpus_tasks[holdout_index]
        
        train_A = np.concatenate([encoded_splits[i][0] for i in range(len(encoded_splits)) if i != holdout_index])
        train_b = np.concatenate([encoded_splits[i][1] for i in range(len(encoded_splits)) if i != holdout_index])

        holdout_A = encoded_splits[holdout_index][0]
        holdout_b = encoded_splits[holdout_index][1]

        model = fit_model(train_A, to_log(train_b), 0, 0.5)

        b_star = from_log(model.predict(holdout_A))
        for holdout_a, b, bs in zip(holdout_A, holdout_b, b_star):
            mp = model_from_features(holdout_a, dim_labels)
            model_pairs[mp].append((bs, b))
            model_pairs["Overall"].append((bs, b))

    for mp, pairs in model_pairs.items():
        mse = 0
        mae = 0
        mean_pred = 0
        mean_gold = 0
        for pred, gold in pairs:
            mse += (pred-gold)**2 / len(pairs)
            mae += abs(pred-gold) / len(pairs)
            mean_pred += pred / len(pairs)
            mean_gold += gold / len(pairs)

        # now we have means, go through again for (co)variances
        var_pred = 0
        var_gold = 0
        cov = 0
        for pred, gold in pairs:
            var_pred += (pred-mean_pred)**2 / len(pairs)
            var_gold += (gold-mean_gold)**2 / len(pairs)
            cov += (pred-mean_pred) * (gold-mean_gold) / len(pairs)
        r = cov / (var_gold*var_pred)**.5
        print(mp)
        print("RMSE", mse**.5)
        print("MAE", mae)
        print("r", r)
        print("r**2", r**2)
        print()

def fit_model(A, b, alpha, ratio):
    scaler = preprocessing.StandardScaler()

    if alpha <= 0.01:
        regression = linear_model.Ridge(alpha=0.01)

    else:
        if ratio <= 0.01:
            regression = linear_model.Ridge(alpha=alpha)
        elif ratio >= 0.99:
            regression = linear_model.Lasso(alpha=alpha, max_iter=5000)
        else:
            regression = linear_model.ElasticNet(alpha=alpha, l1_ratio=ratio)

        
    pipeline = make_pipeline(scaler, regression)
    
    pipeline.fit(A, b)
    return pipeline

def display_significance_test(f1s):
    """
    Use the statsmodel library instead of sklearn, since it does significance testing for us
    Assume we are using no regularization, just OLS
    """
    A, b, dim_labels = encode(f1s)
    # ugly mix of libraries because I don't know statsmodels well
    scaler = preprocessing.StandardScaler()
    A = scaler.fit_transform(A)

    # for intercept
    A2 = sm.add_constant(A)

    est = sm.OLS(to_log(b), A2)
    for i, dim_label in enumerate(dim_labels):
        est.exog_names[i+1] = dim_label
    est2 = est.fit()
    print(est2.summary(alpha=0.002))


def encode_splits(f1s, dim_labels=None):
    if dim_labels is None:
        _, _, dim_labels = encode(f1s)
    splits = {}
    for f1 in f1s:
        corpus_task = f1[1:3]
        if corpus_task not in splits:
            splits[corpus_task] = []
        splits[corpus_task].append(f1)
    encoded_splits = []
    corpus_tasks = []
    for corpus_task, split in splits.items():
        corpus_tasks.append(corpus_task)
        encoded_splits.append(encode(split, dim_labels)[:-1])
    return encoded_splits, dim_labels, corpus_tasks
        
def encode(f1s, dim_labels=None):
    def model_features(model_properties):
        return {
            'CRF': float('CRF' in model_properties),
            'LSTM': float('LSTM' in model_properties),
            'BERT': float('BERT' in model_properties),
            'Feat': float('Feat' in model_properties),
        }
    def task_features(corpus, task):
        feats = {}
        tp = get_task_properties(corpus, task)

        feats['n_spans'] = np.log(tp.n_spans)
        feats['span_length'] = np.log(tp.geom_mean_span_length)
        feats['span_divergence'] = tp.span_divergence
        feats['boundary_divergence'] = tp.boundary_divergence
        return feats

    def prod(a, b, symmetric=True, squares=False):
        c = {}
        for a_feat, a_val in a.items():
            for b_feat, b_val in b.items():
                key = ()
                if isinstance(a_feat, tuple):
                    key += a_feat
                else:
                    key += (a_feat,)
                if isinstance(b_feat, tuple):
                    key += b_feat
                else:
                    key += (b_feat,)
                #key = (a_feat, b_feat)
                if symmetric:
                    key = tuple(sorted(key))
                if squares or len(set(key)) == len(key):
                    c[key] = a_val * b_val
        return c


    if dim_labels is None:
        # find all the dim labels
        dim_labels = set()
        
        for model_properties, corpus, task, f1 in f1s:
            mfeats = model_features(model_properties)
            cfeats = task_features(corpus, task)
            feats = {}
            #feats["dummy"] = 0
            feats.update(mfeats)
            feats.update(cfeats)
            #feats.update(prod(feats, mfeats))
            #feats.update(prod(feats, mfeats))

            dim_labels.update(feats.keys())

        dim_labels = list(dim_labels)

    #calculate all the feat dicts, and make our A matrix and b vector
    A = []
    b = []
    for model_properties, corpus, task, f1 in f1s:
        mfeats = model_features(model_properties)
        cfeats = task_features(corpus, task)
        feats = {}
        #feats["dummy"] = 0
        feats.update(mfeats)
        feats.update(cfeats)
        #feats.update(prod(feats, mfeats))
        #feats.update(prod(feats, mfeats))
        
        feat_vector = [feats.get(k, 0) for k in dim_labels]
        A.append(feat_vector)
        b.append(f1)


    A = np.array(A)
    b = np.array(b)

    return (A, b, dim_labels)
    


def make_plot_data(f1s):
    """
    Exports data that is useful for making a plot
    """
    models = set()
    
    real_f1s = {}
    pred_f1s = {}
    # both are dicts with (model, corpus, task) as keys and average f1 scores as values
    for model, corpus, task, score in f1s:
        models.add(model)
        if (model, corpus, task) not in real_f1s:
            real_f1s[model, corpus, task] = 0
        real_f1s[model, corpus, task] += score / 5
    
    encoded_splits, dim_labels, corpus_tasks = encode_splits(f1s)
    
    # we already know that we just use OLS, so no hyperparameter search
    for holdout_index in range(len(encoded_splits)):
        corpus, task = corpus_tasks[holdout_index]
        
        #holdout_A = encoded_splits[holdout_index][0]
        #holdout_b = encoded_splits[holdout_index][1]
        train_A = np.concatenate([encoded_splits[i][0] for i in range(len(encoded_splits)) if i != holdout_index])
        train_b = np.concatenate([encoded_splits[i][1] for i in range(len(encoded_splits)) if i != holdout_index])

        model = fit_model(train_A, to_log(train_b), 0, 0.5)
        
        for m in models:
            dummy_f1s = [(m, corpus, task, 0)]
            A, _, _ = encode(dummy_f1s, dim_labels)
            res = from_log(model.predict(A))[0]
            pred_f1s[m, corpus, task] = res
    xs = []
    ys = []
    models = []
    corpus_tasks =  []
    for model, corpus, task in real_f1s:
        xs.append(real_f1s[model, corpus, task])
        ys.append(pred_f1s[model, corpus, task])
        models.append(model)
        corpus_tasks.append((corpus, task))
    # write to file so they can be plotted later
    with open("plot_data.pkl", 'wb') as f:
        pickle.dump((xs, ys, models, corpus_tasks), f)

if (__name__ == '__main__'):
    f1s = get_f1s()
    #perform_regression(f1s)
    #detailed_error_data(f1s)
    #make_plot_data(f1s)
    #display_significance_test(f1s)

