import os
import re

from copy import deepcopy

import spacy

from corpus import Corpus, Document, Token

nlp = spacy.load("en_core_web_sm")


dig = re.compile('[0-9]')
def normalize_text(s):
    #such a clean and beautiful interface...
    if s is None:
        return s
    s = s.lower()
    s = re.sub(dig, '0', s)
    s = s.replace("``", '"')
    s = s.replace("''", '"')
    return s

def load_document(txt_path, ann_path, span_types=None, restricted_features=False):
    with open(txt_path, 'r') as f:
        text = f.read()

    tokens = []
    char2tok = {}

    for token in nlp(text):
        l = token.idx
        r = l + len(token)
        for i in range(l, r):
            char2tok[i] = len(tokens)
        assert token.text == text[l:r]
        if restricted_features:
            features = [
                ('word', normalize_text(token.text))
            ]
        else:
            features = [
                ('word', normalize_text(token.text)),
                ('pos', token.pos_),
                ('lemma', token.lemma_),
                ('quote', token.is_quote),
                ('allCaps', token.is_upper),
                ('upperInitial', token.text[0].isupper())
            ]
        tokens.append(Token(token.text, features))
    labels = ['O'] * len(tokens)
    spans = []
    with open(ann_path, 'r') as f:
        for line in f:
            line = line.strip()
            if line.startswith('T'):
                cols = line.split('\t')
                assert len(cols) == 3
                span_type, left, right = cols[1].split()
                if span_types is not None and span_type not in span_types:
                    continue
                left = int(left)
                right = int(right)
                left_token = char2tok[left]
                right -= 1
                while right not in char2tok:
                    right -= 1
                right_token = char2tok[right] + 1
                spans.append((span_type, left_token, right_token))
    # filter nested spans
    filtered_spans = []
    for i, (inner_role, inner_l, inner_r) in enumerate(spans):
        for outer_role, outer_l, outer_r in spans[:i] + spans[i+1:]:
            assert (inner_l, inner_r) != (outer_l, outer_r)
            # completely inside
            if inner_l > outer_l and inner_r < outer_r:
                break
            # shared l
            if inner_l == outer_l and inner_r < outer_r:
                break
            # shared r
            if inner_r == outer_r and inner_l > outer_l:
                break
        else:
            assert not outer_l < inner_l < outer_r
            assert not outer_l < inner_r < outer_r
            filtered_spans.append((inner_role, inner_l, inner_r))

    for span_type, l, r in filtered_spans:
        labels[l] = 'B-' + span_type
        for i in range(l+1, r):
            labels[i] = 'I-' + span_type
    return Document(tokens, labels)

def iter_xvalidation_splits(path, log, span_types=None, restricted_features=False):
    author_documents = {}
    for dirpath, dirnames, filenames in os.walk(path):
        for ann_filename in filenames:
            if ann_filename.endswith('.ann'):
                txt_filename = ann_filename[:-4] + '.txt'
                assert txt_filename in filenames
                ann_path = os.path.join(dirpath, ann_filename)
                txt_path = os.path.join(dirpath, txt_filename)
                author = ann_filename.split('_')[0]
                if author not in author_documents:
                    author_documents[author] = []
                log("Loading document %s" % (ann_filename[:-4]))
                document = load_document(txt_path, ann_path, span_types=span_types, restricted_features=restricted_features)
                author_documents[author].append(document)
    authors = sorted(author_documents.keys())
    for split in range(len(authors)):
        train_authors = authors[:-2]
        dev_author = authors[-2]
        test_author = authors[-1]
        train_documents = []
        for author in train_authors:
            train_documents.extend(author_documents[author])
        dev_documents = list(author_documents[dev_author])
        test_documents = list(author_documents[test_author])

        log("Split %d\n Train authors: %s\n Dev author: %s\n Test author: %s" % (
            split,
            ', '.join(train_authors),
            dev_author,
            test_author
        ))
        yield Corpus(deepcopy(train_documents)), Corpus(deepcopy(dev_documents)), Corpus(deepcopy(test_documents))
        authors = authors[1:] + authors[:1]


def load_corpus(path, span_types=None, restricted_features=False):
    documents = []
    for dirpath, dirnames, filenames in os.walk(path):
        for ann_filename in filenames:
            if ann_filename.endswith('.ann'):
                txt_filename = ann_filename[:-4] + '.txt'
                assert txt_filename in filenames
                ann_path = os.path.join(dirpath, ann_filename)
                txt_path = os.path.join(dirpath, txt_filename)
                documents.append(load_document(txt_path, ann_path, span_types=span_types, restricted_features=restricted_features))
    assert len(documents) > 0
    return Corpus(documents)

if __name__ == '__main__':
    import sys
    corpus = load_corpus(sys.argv[1])
    breakpoint()
