import torch
import torch.nn as nn
import torch.nn.functional as F

from allennlp.modules.seq2seq_encoders.stacked_self_attention import StackedSelfAttentionEncoder

from crf_layer import CRFOutputLayer

class StackedSequenceEncoder(nn.Module):
    def __init__(self, encoders, residual=True):
        super().__init__()
        self.encoders = nn.ModuleList(encoders)
        self.residual = residual

    def forward(self, x):
        for encoder in self.encoders:
            if self.residual:
                x = x + encoder(x)
            else:
                x = encoder(x)
        return x

class NOPSequenceEncoder(nn.Module):
    def __init__(self, n_units):
        super().__init__()
        self.n_units = n_units

    def forward(self, x):
        return x

class TransformerSequenceEncoder(nn.Module):
    def __init__(self, n_units, n_layers=1, n_heads=6, window_size=256):
        super().__init__()
        assert n_units % n_heads == 0
        assert window_size % 4 == 0
        self.n_units = n_units
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.window_size = window_size
        self.transformer = StackedSelfAttentionEncoder(
            input_dim=n_units,
            hidden_dim=n_units,
            projection_dim=n_units,
            feedforward_hidden_dim=4*n_units,
            num_layers=n_layers,
            num_attention_heads=n_heads
        )

    def forward(self, x):
        # x: float[batch_size, sequence_length, self.n_units
        batch_size = x.shape[0]
        sequence_length = x.shape[1]

        if sequence_length <= self.window_size:
            mask = x.new_ones(batch_size, sequence_length, dtype=torch.float)
            #mask.requires_grad = False
            return self.transformer(x, mask)

        output_windows = []
        for window_start in range(0, sequence_length, self.window_size//2):
            input_window = x[:,window_start:window_start+self.window_size]
            window_length = input_window.shape[1]
            # input_window: float32[batch_size, window_length]
            mask = x.new_ones(batch_size, window_length, dtype=torch.float)
            #mask.requires_grad = False
            output_window = self.transformer(input_window, mask)
            # output_window: float32[batch_size, window_length]
            output_windows.append(output_window)
        window_start_index = self.window_size//4
        window_end_index = 3*self.window_size//4
        parts = []
        parts.append(output_windows[0][:,:window_end_index])
        for output_window in output_windows[1:-1]:
            parts.append(output_window[:,window_start_index:window_end_index])
        parts.append(output_windows[-1][:,window_start_index:])
        out = torch.cat(parts, dim=1)
        return out

class LSTMSequenceEncoder(nn.Module):
    def __init__(self, n_units, bidirectional=True, n_layers=2, dropout=0.5):
        super().__init__()
        self.n_units = n_units
        self.n_layers = n_layers
        self.bidirectional = bidirectional
        self.lstm = nn.LSTM(self.n_units, self.n_units, batch_first=True, bidirectional=bidirectional, num_layers=n_layers, dropout=dropout)
        if self.bidirectional:
            self.output = nn.Linear(2*self.n_units, self.n_units)

    def forward(self, x):
        # x: float32[batch_size, sequence_length, self.n_units]
        lstm_out, _ = self.lstm(x)
        if self.bidirectional:
            return self.output(lstm_out)
        else:
            return lstm_out

class SoftmaxOutputLayer(nn.Module):
    def __init__(self, n_units, n_labels):
        super().__init__()
        self.n_units = n_units
        self.n_labels = n_labels
        self.weights = nn.Linear(n_units, n_labels)

    def loss(self, x, y, lengths):
        # x: float32[batch_size, sequence_length, self.n_units]
        # y: int64[batch_size, sequence_length]
        # lengths: int64[batch_size]
        batch_size = x.shape[0]
        sequence_length = x.shape[1]
        x = self.weights(x)
        # x: float32[batch_size, sequence_length, self.n_labels]
        loss = F.cross_entropy(x.view(-1, self.n_labels), y.view(-1), reduction='none').view(batch_size, -1)
        # loss: float32[batch_size, sequence_length]
        length_mask = torch.arange(sequence_length).to(x.device).expand(batch_size, sequence_length) < lengths.unsqueeze(1)
        length_mask = length_mask.float()
        # length_mask: float32[batch_size, sequence_length]
        mean_loss = torch.sum(loss * length_mask) / torch.sum(length_mask)
        return mean_loss

    def forward(self, x):
        # x: float32[batch_size, sequence_length, self.n_units]
        x = self.weights(x)
        # x: float32[batch_size, sequence_length, 3]
        return torch.argmax(x, dim=-1)

class AutoregressiveOutputLayer(nn.Module):
    def __init__(self, n_units, n_labels, n_layers=1):
        super().__init__()
        self.n_units = n_units
        self.n_labels = n_labels
        self.n_layers = n_layers
        self.embedding = nn.Embedding(n_labels, n_units)
        self.lstm = nn.LSTM(n_units, n_units, batch_first=True, num_layers=n_layers)
        self.output = nn.Linear(n_units, n_labels)

    def forward(self, x, temperature=0):
        # x: float32[batch_size, sequence_length, n_units]
        batch_size = x.shape[0]
        sequence_length = x.shape[1]
        h = c = x.new_zeros(self.n_layers, batch_size, self.n_units)
        embedded_outputs = x.new_zeros(batch_size, self.n_units)
        # embedded_outputs: float32[batch_size, n_units]
        output_sequence = []
        for i in range(sequence_length):
            lstm_input = x[:,i,:] + embedded_outputs
            # lstm_input: float32[batch_size, self.n_units]
            lstm_input = lstm_input.view(batch_size, 1, self.n_units)
            # lstm_input: float32[batch_size, 1, self.n_units]
            lstm_out, (h, c) = self.lstm(lstm_input, (h, c))
            lstm_out = lstm_out.view(batch_size, self.n_units)
            # lstm_out: float32[batch_size, self.n_units]
            logits = self.output(lstm_out)
            # logits: float32[batch_size, 3]
            if temperature > 0:
                logits = logits / temperature
                p = torch.softmax(logits, -1)
                # p: float32[batch_size, 3]
                samples = torch.multinomial(p, 1).view(batch_size)
                # samples: int64[batch_size]
            else:
                samples = torch.argmax(logits, dim=-1)
            output_sequence.append(samples)
            embedded_outputs = self.embedding(samples)
            # embedded_outputs: float32[batch_size, n_units]
        output_sequence = torch.stack(output_sequence, dim=-1)
        # output_sequence: int64[batch_size, sequence_length]
        return output_sequence

    def loss(self, x, y, lengths):
        # x: float32[batch_size, sequence_length, self.n_units]
        # y: int64[batch_size, sequence_length]
        batch_size = x.shape[0]
        sequence_length = x.shape[1]
        y_embedded = self.embedding(y[:,:-1])
        # y_embedded: float32[batch_size, sequence_length-1, self.n_units]
        y_embedded = F.pad(y_embedded, (0,0,1,0))
        # y_embedded: float32[batch_size, sequence_length, self.n_units]
        lstm_out, _ = self.lstm(x + y_embedded)
        # lstm_out: float32[batch_size, sequence_length, n_units]
        logits = self.output(lstm_out)
        # logits: float32[batch_size, sequence_length, 3]
        loss = F.cross_entropy(logits.view(-1, 3), y.view(-1), reduction='none').view(batch_size, -1)
        # loss: float32[batch_size, sequence_length]
        length_mask = torch.arange(sequence_length).to(x.device).expand(batch_size, sequence_length) < lengths.unsqueeze(1)
        length_mask = length_mask.float()
        # length_mask: float32[batch_size, sequence_length]
        mean_loss = torch.sum(loss * length_mask) / torch.sum(length_mask)
        return mean_loss


class SequenceLabeler(nn.Module):
    def __init__(self, model_type, vocab_size, n_units, n_labels):
        super().__init__()
        self.vocab_size = vocab_size
        self.n_units = n_units
        self.n_labels = n_labels
        self.model_type = model_type
        self.embedding = nn.EmbeddingBag(self.vocab_size, self.n_units)
        if self.model_type == 'baseline':
            self.sequence_encoder = lambda x:x
            self.output_layer = SoftmaxOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'lstm':
            self.sequence_encoder = LSTMSequenceEncoder(self.n_units)
            self.output_layer = SoftmaxOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'crf':
            self.sequence_encoder = lambda x:x
            self.output_layer = CRFOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'lstm-crf':
            self.sequence_encoder = LSTMSequenceEncoder(self.n_units)
            self.output_layer = CRFOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'lstm-autoregressive':
            self.sequence_encoder = LSTMSequenceEncoder(self.n_units)
            self.output_layer = AutoregressiveOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'lstm-xf-crf':
            self.sequence_encoder = StackedSequenceEncoder([
                LSTMSequenceEncoder(self.n_units),
                TransformerSequenceEncoder(self.n_units)
            ])
            self.output_layer = CRFOutputLayer(self.n_units, self.n_labels)
        elif self.model_type == 'xf-lstm-crf':
            self.sequence_encoder = StackedSequenceEncoder([
                TransformerSequenceEncoder(self.n_units),
                LSTMSequenceEncoder(self.n_units)
            ])
            self.output_layer = CRFOutputLayer(self.n_units, self.n_labels)
        else:
            raise ValueError('Unknown model type: %s' % model_type)
    
    def forward(self, x):
        if isinstance(x, tuple):
            indices, offsets = x
            # indices: int64[n_features]
            # offsets: int64[batch_size, sequence_length]
            batch_size, sequence_length = offsets.shape
            # offset_addition: int64[batch_size, 1]
            embedded = self.embedding(indices, offsets.view(-1)).view(batch_size, sequence_length, self.n_units)
            # embedded: float32[batch_size, sequence_length, self.n_units]
        else:
            # x: int64[batch_size, sequence_length]
            batch_size, sequence_length = x.shape
            offsets = torch.arange(batch_size*sequence_length).to(x.device)
            embedded = self.embedding(x.view(-1), offsets).view(batch_size, sequence_length, self.n_units)
            # embedded: float32[batch_size, sequence_length, self.n_units]

        sequence_encoded = embedded + self.sequence_encoder(embedded)
        # sequence_encoded: float32[batch_size, sequence_length, self.n_units]
        out = self.output_layer(sequence_encoded)
        # out: int64[batch_size, sequence_length]
        return out
    
    def loss(self, x, y, lengths):
        if isinstance(x, tuple):
            indices, offsets = x
            # indices: int64[n_features]
            # offsets: int64[batch_size, sequence_length]
            batch_size, sequence_length = offsets.shape
            # offset_addition: int64[batch_size, 1]
            embedded = self.embedding(indices, offsets.view(-1)).view(batch_size, sequence_length, self.n_units)
            # embedded: float32[batch_size, sequence_length, self.n_units]
        else:
            # x: int64[batch_size, sequence_length]
            batch_size, sequence_length = x.shape
            offsets = torch.arange(batch_size*sequence_length).to(x.device)
            embedded = self.embedding(x.view(-1), offsets).view(batch_size, sequence_length, self.n_units)
            # embedded: float32[batch_size, sequence_length, self.n_units]

        # y: int64[batch_size, sequence_length]
        # lengths: int64[batch_size]
        # embedded: float32[batch_size, sequence_length, self.n_units]
        sequence_encoded = embedded + self.sequence_encoder(embedded)
        # sequence_encoded: float32[batch_size, sequence_length, self.n_units]
        return self.output_layer.loss(sequence_encoded, y, lengths)

