import re
import pickle
from collections import defaultdict

import numpy as np

from evaluation import spanset

parc_path = "..."
ontonotes_path = "..."
chemdner_path = "..."
riqua_path = "..."
conll2k_path = "..."


dig = re.compile('[0-9]')
def normalize_text(s):
    #such a clean and beautiful interface...
    if s is None:
        return s
    s = s.lower()
    s = re.sub(dig, '0', s)
    s = s.replace("``", '"')
    s = s.replace("''", '"')
    return s

def normalize(counts):
    dist = dict(counts)
    sm = 0
    for word in dist:
        sm += dist[word]
    for word in dist:
        dist[word] /= sm
    return dist

class TaskProperties:
    def __init__(self):
        self.n_spans = 0
        self.inside_counts = defaultdict(int)
        self.boundary_counts = defaultdict(int)
        self.span_lengths = []

def kl(p, q):
    sm = 0
    for word, p_word in p.items():
        sm += p_word * np.log(p_word / q[word])
    return sm

def compute_task_properties(corpus):
    unigram_counts = defaultdict(int)
    task_properties = defaultdict(TaskProperties)
    for document in corpus.documents:
        for token in document.tokens:
            word = normalize_text(token.text)
            unigram_counts[word] += 1

        for span_type, start, end in spanset(document.labels):
            tp = task_properties[span_type]
            tp.n_spans += 1
            tp.span_lengths.append(end-start)
            for token in document.tokens[start:end]:
                word = normalize_text(token.text)
                tp.inside_counts[word] += 1
            if start > 0:
                token = document.tokens[start-1]
                word = normalize_text(token.text)
                tp.boundary_counts[word] += 1
            if end < len(document.tokens):
                token = document.tokens[end]
                word = normalize_text(token.text)
                tp.boundary_counts[word] += 1
    unigram_dist = normalize(unigram_counts)
    for task, tp in task_properties.items():
        #print(task)
        #print("# spans:", tp.n_spans)
        span_lengths = np.array(tp.span_lengths)
        tp.arit_mean_span_length = np.mean(span_lengths)
        tp.geom_mean_span_length = np.exp(np.mean(np.log(span_lengths)))
        #print("arit. mean length:", np.mean(span_lengths))
        #print("geom. mean length:", np.exp(np.mean(np.log(span_lengths))))
        inside_dist = normalize(tp.inside_counts)
        tp.span_divergence = kl(inside_dist, unigram_dist)
        #print("span divergence:", kl(inside_dist, unigram_dist))
        boundary_dist = normalize(tp.boundary_counts)
        tp.boundary_divergence = kl(boundary_dist, unigram_dist)
        #print("boundary divergence:", kl(boundary_dist, unigram_dist))
    return dict(task_properties)


def load_corpus(corpus_name):
    if corpus_name == 'parc':
        import parc_loader
        corpus = parc_loader.load_corpus(parc_path)
    if corpus_name == "ontonotes":
        import ontonotes_loader
        corpus = ontonotes_loader.load_corpus(ontonotes_path)
    if corpus_name == "riqua":
        import riqua_loader
        corpus = riqua_loader.load_corpus(riqua_path)
    if corpus_name == "conll2k":
        import conll2k_loader
        corpus = conll2k_loader.load_corpus(conll2k_path)
    if corpus_name == "chemdner":
        import chemdner_loader
        corpus = chemdner_loader.load_corpus(chemdner_path)

    assert len(corpus.documents) > 0
    return corpus

try:
    with open("cached_properties.pkl", 'rb') as f:
        cached_properties = pickle.load(f)
except FileNotFoundError:
    cached_properties = {}
def get_task_properties(corpus_name, task_name):
    if corpus_name not in cached_properties:
        corpus = load_corpus(corpus_name)
        cached_properties[corpus_name] = compute_task_properties(corpus)
        if corpus_name == "riqua":
            for task_name in cached_properties[corpus_name]:
                cached_properties[corpus_name][task_name].n_spans *= 5/6
        with open("cached_properties.pkl", 'wb') as f:
            pickle.dump(cached_properties, f)
    return cached_properties[corpus_name][task_name]

def include_task(corpus_name, task_name):
    if corpus_name == 'riqua' and task_name == "Entity":
        return False
    if corpus_name == 'parc' and task_name == 'source':
        return False
    return get_task_properties(corpus_name, task_name).n_spans > 100
